////////////////////////////////////////////////////////////////////////////////
// Simple OpenCV DNN module warper for Matlab/Octave
//
// Author: A. Varfolomieiev
// NTUU "Igor Sikorsky Kyiv Polytechnic Institute"
//

#include <vector>

#include "opencv2/core.hpp"
#include "opencv2/dnn.hpp"
#include "opencv2/highgui.hpp"

#include "mex.h"

#ifndef PATH_MAX
#define PATH_MAX 256
#endif

#ifdef _WIN32
#define strcasecmp _strcmpi
#endif

using namespace cv;
using namespace dnn;
using namespace std;

////////////////////////////////////////////////////////////////////////////////
// string case insensitive comparison
int strcmpi(const char *str1, const char *str2)
{
	int  r = 0;
	char *s1 = (char *)str1, *s2 = (char *)str2;

	do {
		r = tolower(*s2++) - tolower(*s1++);
	} while (r == 0 && *s1);

	return r;
}

const mxArray *getParamByKey(int nrhs, const mxArray *prhs[], const char *key, mxClassID checkType)
{
	int  i;
	char str[64];

	for (i = 0; i < nrhs; i++) {
		if (mxGetClassID(prhs[i]) == mxCHAR_CLASS) {
			mxGetString(prhs[i], str, 63);
			if (0 == strcasecmp(str, key) && i < nrhs) {
				if (mxGetClassID(prhs[i + 1]) == checkType) {
					return prhs[i + 1];
				}
				else {
					mexWarnMsgIdAndTxt("MATLAB:narginchk:InconsistentDataType",
					                   "Incorrect data type of parameter after the key: \"%s\"", str);
				}
			}
		}
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Conversion between OpenCV Mat and Matlab mxArray
void loadCvImageFromMatlab(Mat &cvMat, const mxArray *mxMat)
{
	int	      i, j, k,
	          rows, cols, cn;
	uchar     *matData;
	const int *size;

	if (mxGetNumberOfDimensions(mxMat) > 3) {
		mexErrMsgTxt("Only images with up to 3 dimensions are supported.");
	}

	size = mxGetDimensions(mxMat);

	rows = size[0];
	cols = (mxGetNumberOfDimensions(mxMat) == 1) ? 1 : size[1];

	if (mxGetNumberOfDimensions(mxMat) == 3) {
		cn = size[2];
	}
	if (cn != 1 && cn != 3 && cn != 4) {
		mexErrMsgTxt("Only images with 1-, 3- and 4-channels are supported.");
	}
	
	cvMat.create(rows, cols, CV_MAKETYPE(CV_8U, cn));

	matData = (uchar *)mxGetData(mxMat);
	for (k = 0; k < cn; k++) {				// channels
		for (i = 0; i < rows; i++) {		// rows
			for (j = 0; j < cols; j++) {	// cols
				cvMat.data[k + j * cn + i * cvMat.step1(0)] = matData[i + j * rows + k * (rows * cols)];
			}
		}
	}
}

void returnCvMatToMatlab(mxArray **mxMat, Mat &cvMat)
{
	float  *matData,
	       *cvData;
	mwSize size[3];
	size_t step[2],
	       dims;

	dims = (cvMat.dims > 3) ? 3 : cvMat.dims;
	if (dims <= 2) {
		size[0] = cvMat.rows;		// rows
		size[1] = cvMat.cols;		// cols
		size[2] = 1;				// channels
		step[0] = cvMat.step1();
		step[1] = 0;
	}
	else {
		size[0] = cvMat.size[2];	// rows
		size[1] = cvMat.size[3];	// cols
		size[2] = cvMat.size[1];	// channels
		step[0] = cvMat.step1(2);
		step[1] = cvMat.step1(1);
	}
	(*mxMat) = mxCreateNumericArray(dims, size, mxSINGLE_CLASS, mxREAL);
	
	matData = (float *)mxGetData(*mxMat);
	cvData  = (float *)cvMat.data;
	for (size_t k = 0; k < size[2]; k++) {			// channels
		for (size_t i = 0; i < size[0]; i++) {		// rows
			for (size_t j = 0; j < size[1]; j++) {	// cols
				matData[i + j * size[0] + k * size[0] * size[1]] = cvData[j + i * step[0] + k * step[1]];
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	static Net            net;
	static Mat            img,
	                      input;	// Input blob
	static vector<Mat>    output;	// Output of layer, which will be specified
	double                scale = 1.0;
	static Scalar         mean;
	static Size           size;
	static vector<String> names;
	
	////////////////////////
	// Network loading
	if (nrhs >= 2 && mxGetClassID(prhs[0]) == mxCHAR_CLASS) {
		if (mxGetClassID(prhs[1]) != mxCHAR_CLASS) {
			mexErrMsgTxt("Second parameter should be a string that specifies file name of a network weights.");
		}

		// Try to load net
		try {
			char config[PATH_MAX],
			     weights[PATH_MAX];

			mxGetString(prhs[0], config, PATH_MAX - 1);
			mxGetString(prhs[1], weights, PATH_MAX - 1);

			net = readNet(weights, config);

			// Try to set target
			if (nrhs == 3 && mxGetClassID(prhs[2]) == mxCHAR_CLASS) {
				char target[16];
				mxGetString(prhs[2], target, 16 - 1);
				if (0 == strcasecmp(target, "cpu"))
					net.setPreferableTarget(DNN_TARGET_CPU);
				if (0 == strcasecmp(target, "cl"))
					net.setPreferableTarget(DNN_TARGET_OPENCL);
				if (0 == strcasecmp(target, "cl16"))
					net.setPreferableTarget(DNN_TARGET_OPENCL_FP16);
			}

			// Get layer names
			names = net.getLayerNames();
			const char **str = (const char **)malloc(names.size() * sizeof(char *));
			for (size_t i = 0; i < names.size(); i++) {
				str[i] = names[i].c_str();
			}
			plhs[0] = mxCreateCharMatrixFromStrings(names.size(), str);
			free(str);
		}
		catch (exception &ex) {
			mexErrMsgTxt(ex.what());
		}
		return;
	}
	
	////////////////////////
	// Do network inference
	if (nrhs >= 2 && mxGetClassID(prhs[0]) == mxUINT8_CLASS) {
		if (mxGetClassID(prhs[1]) != mxCHAR_CLASS &&
			mxGetClassID(prhs[1]) != mxDOUBLE_CLASS && 
			!mxIsCell(prhs[1])) {
			mexErrMsgTxt("Second parameter should be a string, number or "
				         "cell array that specifies a name(s) or id(s) "
			             "of layer(s), which have to be computed.");
		}

		// Try do inference
		if (net.empty()) {
			mexErrMsgTxt("Please, load network parameters before inference.");
		}
		try {
			vector<String> layers;
			const mxArray  *param;

			if (mxIsCell(prhs[1])) {
				// Second argument is cell array
				mwSize cellsNum = mxGetNumberOfElements(prhs[1]);
				for (mwIndex index = 0; index < cellsNum; index++) {
					const mxArray *cell = mxGetCell(prhs[1], index);

					if (mxGetClassID(cell) == mxDOUBLE_CLASS) {	// Cell is number
						size_t id = (size_t)fabs(floor(*(double *)mxGetData(cell)) - 1);
						if (id < names.size()) {
							layers.push_back(names[id].c_str());
						}
					}
					else {										
						char layer[PATH_MAX] = "";
						mxGetString(cell, layer, PATH_MAX - 1);	// Get layer name
						layers.push_back(layer);
					}
				}
			}
			else {
				//  Second argument is single element
				if (mxGetClassID(prhs[1]) == mxDOUBLE_CLASS) {  // Single number given
					size_t id = (size_t)fabs(floor(*(double *)mxGetData(prhs[1])) - 1);
					if (id < names.size()) {
						layers.push_back(names[id].c_str());
					}
				}
				else {
					char layer[PATH_MAX] = "";
					mxGetString(prhs[1], layer, PATH_MAX - 1);  // Get layer name
					layers.push_back(layer);
				}
			}
			
			// Parameter parsing
			if (param = getParamByKey(nrhs, prhs, "scale", mxDOUBLE_CLASS)) {
				scale = *(double *)mxGetData(param);
				scale = max(min(scale, 1.0), 0.0);		// clamp scale
			}
			else {
				scale = 1;
			}
			if (param = getParamByKey(nrhs, prhs, "size", mxDOUBLE_CLASS)) {
				size.width  = (int)max(((double *)mxGetData(param))[0], 0.0);
				size.height = (int)max(((double *)mxGetData(param))[1], 0.0);
			}
			else {
				size = Size();
			}
			if (param = getParamByKey(nrhs, prhs, "mean", mxDOUBLE_CLASS)) {
				if (mxGetN(param) >= 1)
					mean.val[0] = ((double *)mxGetData(param))[0];
				if (mxGetN(param) >= 3)
					mean.val[1] = ((double *)mxGetData(param))[1];
					mean.val[2] = ((double *)mxGetData(param))[2];
				if (mxGetN(param) == 4 )
					mean.val[3] = ((double *)mxGetData(param))[3];
			}
			else {
				mean = Scalar();
			}
			
			/////////////////////////
			// Inference
			loadCvImageFromMatlab(img, prhs[0]);

			input = blobFromImage(img, scale, size, mean, true);
			net.setInput(input);
			net.forward(output, layers);	// Do forward pass

			/////////////////////////
			// Returning result
			if (1 == output.size()) {
				// Single blob output
				returnCvMatToMatlab(&plhs[0], output[0]);
			}
			else {
				// Multiple outputs
				const mwSize dims[1] = { output.size() };
				plhs[0] = mxCreateCellArray(1, dims);
				for (size_t i = 0; i < output.size(); i++) {
					mxArray *outCell;
					returnCvMatToMatlab(&outCell, output[i]);
					mxSetCell(plhs[0], (mwIndex)i, outCell);
				}
			}

		}
		catch (exception &ex) {
			mexErrMsgTxt(ex.what());
		}
		return;
	}
}
